/*
 * mmfd - fd-based secure I/O functions
 * Copyright (C) 2000-2001  Matthew Mondor
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Original author: Matthew Mondor, mmondor@linuxguru.net
 */

/* Provides high-level functions from low-level io functions.
 * Safe functions for reading and writing to filedescriptors (int).
 * Useful for TCP/IP sockets IO.
 * (Matthew Mondor)
 */




/* HEADERS */

#include "mmfd.h"




/* FUNCTIONS */

/* Waits ms milliseconds for data to be available to read from fd.
 * Returns -1 on error/connection lost, 1 if bytes are available, or
 * 0 on timeout
 */
BYTE fdavail(int fd, int ms)
{
	int selr;
	struct pollfd polldata[] = {
		{fd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0}
	};

	selr = poll(polldata, 1, ms);

	if(selr > 0) {
		selr = polldata[0].revents;

		if(selr & POLLERR || selr & POLLHUP || selr & POLLNVAL)
			return (-1);

		if(selr & POLLIN || selr & POLLPRI)
			return (1);
	}

	return (0);
}


/* Secure fgets to use for tcp communications. Returns -1 on timeout,
 * -2 on error, otherwise it returns the lenght of the string read
 */
LONG fdgets(BYTE * buf, LONG max, int fd, int ms)
{
	LONG len = 0;
	BYTE c = 0;
	BYTE tmp;
	LONG err = 0;

	if(fd && buf && max) {

		*buf = 0;
		while(len < max) {
			if((tmp = fdavail(fd, ms)) == 1) {
				if((read(fd, &c, 1)) < 1) {
					err = -2;
					break;
				}						  /* Error */
				if(c == '\n')
					break;
				buf[len] = c;
				len++;
			} else {
				if(tmp == -1)
					err = -2;		  /* Error */
				else
					err = -1;		  /* Timeout */
				break;
			}
		}

		if(len > 0)
			if(buf[len - 1] == '\r')
				len--;
		buf[len] = 0;
		if(err)
			len = err;

	}

	return (len);
}


/* Like puts, but to fd. Returns len of string or -1 on error */
LONG fdputs(int fd, BYTE * buf)
{
	LONG len = 0;

	if(buf && fd)
		len = write(fd, buf, strlen(buf));

	return (len);
}


/* Like fprintf, but to fd. Returns len of string or -1 on error */
LONG fdprintf(int fd, BYTE * fmt, ...)
{
	BYTE fdpfb[1024];
	va_list arg_ptr;

	*fdpfb = 0;
	va_start(arg_ptr, fmt);
	vsnprintf(fdpfb, 1023, fmt, arg_ptr);
	va_end(arg_ptr);

	return (write(fd, fdpfb, strlen(fdpfb)));
}


/* This is a shorter version of fdgets, must be supplied a buffer of
 * at least 1001 characters (1024 recommended). It will timeout after
 * a one minute stalled connection.
 */
LONG getline(BYTE * buf, int fd)
{
	return (fdgets(buf, 1000, fd, 60000));
}

/*
	vim:ts=3:sw=3:cindent:
*/
