/*
 * mmlist.c - doubly linked lists library
 * Copyright (C) 1991-2001  Matthew Mondor
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Original author: Matthew Mondor, mmondor@linuxguru.net
 */




/* HEADERS */

#include "mmlist.h"
#include <stdlib.h>




/* FUNCTIONS */

/* Creates a list in memory, with no lnodes in it, returns null on error
 * This is necessary to create a list.
 */
list *createlist(LONG nodelen)
{
	list *lst;

	if((lst = (list *) calloc(1, sizeof(list)))) {
		lst->l_lnodelen = nodelen;
		return (lst);
	}

	return (NULL);
}


/* Frees a list from memory */
list *freelist(list * lst)
{
	lnode *node, *tmp;

	if(lst) {
		node = lst->l_toplnode;
		if(node) {
			/* Free the whole list */
			while(node) {
				tmp = node->ln_next;
				free(node);
				node = tmp;
			}
		}
		free(lst);
	}

	return (NULL);
}


/* Appends an empty lnode to the list, at the end of all other
 * lnodes. The lnode can be initialised with values afterwards.
 * The new lnode is the new current lnode of the list.
 */
lnode *appendnode(list * lst)
{
	lnode *node, *tmp;

	if((node = (lnode *) calloc(1, lst->l_lnodelen))) {

		/* Link newly allocated lnode to the last lnode in the list */
		if((tmp = lst->l_bottomlnode)) {
			/* There are other lnode(s), link it to last and make last */
			tmp->ln_next = node;
			node->ln_prev = tmp;
			lst->l_bottomlnode = node;
		} else
			/* Only lnode, so only link to list */
			lst->l_bottomlnode = lst->l_toplnode = node;

		/* Increase number of lnodes */
		lst->l_lnodes++;

		/* Make new lnode current */
		lst->l_currentlnode = node;
		return (node);

	}

	return (NULL);
}


/* Inserts an empty lnode to the list, just before the current record
 * position. Again it must be set to values afterwards. The new lnode
 * becomes the current lnode of the list.
 */
lnode *insertnode(list * lst)
{
	lnode *node, *prev, *next;

	if((node = lst->l_currentlnode)) {

		/* Remember link pointers of current record */
		prev = node->ln_prev;
		next = node;				  /* There is always a next */
		/* Allocate new record */
		if(!(node = (lnode *) calloc(1, lst->l_lnodelen)))
			return (NULL);
		/* Increase record counter */
		lst->l_lnodes++;

		/* Re-link database/nextrecord pointers,
		  	there is always a record after
		 */
		node->ln_next = next;
		next->ln_prev = node;
		/* Adjust new current position */
		lst->l_currentlnode = node;
		if(prev) {
			/* There are records before */
			prev->ln_next = node;
			node->ln_prev = prev;
		} else {
			/* There are no records before */
			lst->l_toplnode = node;
			node->ln_prev = NULL;
		}

		return (node);
	}

	return (NULL);
}



/* Deletes the lnode at the current position in the list file.
 * The returned pointer is the updated current lnode position, pointing
 * at the same logical number: the lnode just after the one we deleted.
 */
lnode *deletenode(list * lst)
{
	lnode *node, *prev, *next;

	if((node = lst->l_currentlnode)) {

		/* Remember link pointers of current lnode */
		prev = node->ln_prev;
		next = node->ln_next;
		/* Free current lnode */
		free(node);
		/* Decrease lnode counter */
		lst->l_lnodes--;

		/* Re-link list/nextlnode pointers */
		if(prev) {
			/* There are lnodes before.
			 * If there are lnodes after, link to last lnode
			 */
			if(next)
				next->ln_prev = prev;
			else {
				/* There are no lnodes after */
				prev->ln_next = NULL;
				lst->l_bottomlnode = prev;
			}
		} else {
			/* There are no lnodes before */
			if(next) {
				next->ln_prev = NULL;
				lst->l_toplnode = next;
			} else
				lst->l_toplnode = NULL;
		}
		if(next) {
			/* There are lnodes after.
			 * If there are lnodes before, link to next lnode
			 */
			if(prev)
				prev->ln_next = next;
			else {
				/* There are no lnodes before */
				next->ln_prev = NULL;
				lst->l_toplnode = next;
			}
		} else {
			/* There are no lnodes after */
			if(prev) {
				prev->ln_next = NULL;
				lst->l_bottomlnode = prev;
			} else
				lst->l_bottomlnode = NULL;
		}

		/* Set current lnode to next lnode (may be null) */
		lst->l_currentlnode = next;

		return (next);
	}

	return (NULL);
}


/* Returns pointer to current node of list */
lnode *currentnode(list * lst)
{
	return (lst->l_currentlnode);
}


/* Increases the current pointer to next lnode and returns pointer */
lnode *nextnode(list * lst)
{
	lnode *node;

	if((node = lst->l_currentlnode))
		node = node->ln_next;
	if(node) {
		lst->l_currentlnode = node;
		return (node);
	}

	return (NULL);
}


/* Decreases the current pointer to previous lnode and returns pointer */
lnode *previousnode(list * lst)
{
	lnode *node;

	if((node = lst->l_currentlnode))
		node = node->ln_prev;
	if(node) {
		lst->l_currentlnode = node;
		return (node);
	}

	return (NULL);
}


/* Jumps to first lnode in list, returns new lnode pointer */
lnode *topnode(list * lst)
{
	lst->l_currentlnode = lst->l_toplnode;
	return (lst->l_currentlnode);
}


/* Jumps to last lnode of list, returns new lnode pointer */
lnode *bottomnode(list * lst)
{
	lst->l_currentlnode = lst->l_bottomlnode;
	return (lst->l_currentlnode);
}


/* Returns the total number of lnodes into a list */
ULONG totalnodes(list * lst)
{
	return (lst->l_lnodes);
}


/* Lets the programmer to decide which lnode must be current.
 * Beware! It must be a valid lnode of the list specified!
 */
VOID makecurrent(list * lst, lnode * node)
{
	lst->l_currentlnode = node;
}


/* This function scans the list from the top, until it reaches the specified
 * node number. This is somewhat slow on large lists since it requires to
 * scan the whole list until it reaches the end. Arrays are faster for
 * absolute indexing than linked lists. Returns pointer to node it scans to.
 */
lnode *scantonode(list * lst, ULONG noden)
{
	lnode *node;
	ULONG cnt = noden;

	node = lst->l_toplnode;
	cnt--;

	while(cnt && node) {
		node = node->ln_next;
		cnt--;
	}

	lst->l_currentlnode = node;
	return (node);
}

/*
	vim:ts=3:sw=3:cindent:
*/
